structure KKigyo :> KKigyo =
    (* dp03s nhf v1.0 -- 2003. prilis 11. Hank Dvid *)
struct

(* A bemenetet, ill. a kimenetet a terminlra irnythatjuk Unix alatt
   a /dev/stdin, ill. a /dev/stdout, MS-DOS alatt mindkt esetben a con
   llomnynv megadsval.
*)

exception KKigyo of string;

(* helyes_feladvany kt = A Kigyo.kigyo fggvny els paramtere
      mindenkppen kielgti ezt a prediktumot.
 *)
fun helyes_feladvany (Meret, Kpont, Vpont, RefPL) = 
    let 
	val (MaxS, MaxO) = Meret 
	fun helyes_pont (Sor, Oszlop) = 
	      1 <= Sor andalso 1 <= Oszlop andalso
	      Sor <= MaxS andalso Oszlop <= MaxO
	fun helyes_ref_pont (Sor, Oszlop, Ertek)  =
	      helyes_pont (Sor, Oszlop) andalso
	      0 <= Ertek andalso Ertek <= 7;
	fun tavolsag ((S1, O1), (S2, O2)) = 
	      Math.sqrt(Math.pow(real(S1)-real(S2),2.0)+
			Math.pow(real(O1)-real(O2),2.0))

	fun lexikografikusan_hasonlit ((x1,y1,_), (x2,y2,_)) = 
	      if      x1 < x2 orelse x1 = x2 andalso y1 < y2 then LESS
	      else if x1 > x2 orelse x1 = x2 andalso y1 > y2 then GREATER
	      else EQUAL 
	fun szomszedok_kulonbozok (rp1::(rps as rp2::_)) = 
	      lexikografikusan_hasonlit (rp1, rp2) <> EQUAL
	      andalso szomszedok_kulonbozok rps
	  | szomszedok_kulonbozok _ = true
    in 
	(* a tbla mretezse, a kezd- s vgpontok megadsa helyes: *)
	1 <= MaxS andalso 1 <= MaxO andalso
	helyes_pont Kpont andalso
	helyes_pont Vpont andalso
	(* a kezd- s vgpont is csak a sarkukkal rintkezhetnek: *)
    	tavolsag(Kpont, Vpont) > 1.0 andalso
	(* a RefPL lista minden eleme helyes: *) 
        (List.all (fn x => x) (map helyes_ref_pont RefPL)) andalso 
	(* a RefPL lista lexikografikusan rendezett: *) 
	Listsort.sorted lexikografikusan_hasonlit RefPL andalso 
	(* A RefPL lista minden eleme klnbz (Sor,Oszlop)  
	 koordintra vonatkozik (a rendezettsg miatt elegend 
	 ezt a szomszdos elemekre kiktni): *) 
	szomszedok_kulonbozok RefPL 
    end

(* kigyo_be f = Az f szveges llomnybl beolvasott adatokbl
      kpzett kgytbla.
   kigyo_be : string -> TKigyo.kigyotabla
 *)
fun kigyo_be f =
    let
	fun par [a,b] = (a,b)
	  | par _ =
	    raise KKigyo("Szintaktikusan hibs adat a(z) " ^ f ^ " llomnyban.")

	fun harmas [a,b,c] = (a,b,c)
	  | harmas _ =
	    raise KKigyo("Szintaktikusan hibs adat a(z) " ^ f ^ " llomnyban.")

	fun fuzerbol l2n = l2n o
	                   (List.mapPartial Int.fromString) o
			   (String.tokens Char.isSpace)

	fun refpontok_be s =
	    if TextIO.endOfStream s then []
	    else (fuzerbol harmas) (TextIO.inputLine s) :: refpontok_be s

	fun compare ((y1,x1,_), (y2,x2,_)) =
	    if      y1 > y2 then GREATER
	    else if y1 < y2 then LESS
	    else if x1 > x2 then GREATER
	    else if x1 < x2 then LESS
	    else raise KKigyo("Ismtld rcspont a(z) " ^ f ^ " llomnyban.")

	val istr      = TextIO.openIn f
	val meret     = (fuzerbol par) (TextIO.inputLine istr)
	val kezdo     = (fuzerbol par) (TextIO.inputLine istr)
	val veg       = (fuzerbol par) (TextIO.inputLine istr)
	val refpontok = Listsort.sort compare (refpontok_be istr)
	val tabla     = (meret, kezdo, veg, refpontok) before TextIO.closeIn istr
    in
	if helyes_feladvany tabla then tabla
	else raise KKigyo("Hibs kgytbla a(z) " ^ f ^ " llomnyban.")
    end

(* kigyorajz (kt, kv) = a kv kgyvonalnak megfelel kigyrajz a kt
      kgytbln
   kigyorajz : TKigyo.kigyotabla * TKigyo.kigyovonal -> string
 *)
fun kigyorajz (((m,sz), (ks,ko), (vs,vo), rpk), kv) =
    let
	fun paircmp ((a,b), (c,d)) =
	    case Int.compare(a,c) of
		EQUAL => Int.compare(b,d)
	      | other => other

	fun insert ((S,O,c), bmap) = Binarymap.insert(bmap, (S,O), c)

	val pk = Binarymap.mkDict paircmp
	val pk = foldl insert pk (map (fn (S,O) => (S,O,#"@")) kv)
	val pk = foldl insert pk (map (fn (S,O,v) => (S,O,chr(ord #"0" + v))) rpk)
	val pk = insert ((ks,ko,#"K"), pk)
	val pk = insert ((vs,vo,#"V"), pk)

	val tab   = String.concat o List.tabulate
	val vonal = tab (sz, fn _ => "+---") ^ "+\n"

	fun cellarajz S O = case Binarymap.peek(pk, (S+1,O+1)) of
	                        SOME c => " " ^ str c ^ " |"
			      | NONE   => "   |"

	fun sorrajz i = "|" ^ tab (sz, cellarajz i) ^ "\n" ^ vonal
    in
	vonal ^ tab (m, sorrajz)
    end

(* kigyorajzok (kt, kvk) = a kvk megoldslistnak megfelel kigyrajzok a
      kt kgytbln, res sorokkal elvlasztva
   kigyorajzok : TKigyo.kigyotabla * TKigyo.kigyovonal list -> string
 *)
fun kigyorajzok (kt, kvk) =
    foldl (fn (kv,s) => s ^ kigyorajz(kt, kv) ^ "\n") "" kvk

(* kigyo_ki(m, kt, kvk) = A kt kgytbla kvk megoldslistjt olvashat
      formban kirja az m szveges llomnyba.
   kigyo_ki : string * TKigyo.kigyotabla * TKigyo.kigyovonal list -> unit
 *)
fun kigyo_ki(m, kt, kvk) =
      let val ostr = TextIO.openOut m
	  val _ = TextIO.output(ostr, kigyorajzok(kt, kvk))
      in
	 TextIO.closeOut ostr
      end
  
(* megold(f, m) = Beolvas egy feladvnyt az f szveges llomnybl s
      kirja az sszes megoldst az m szveges llomnyba. Ehhez
      termszetesen szksge van a Kigyo.kigyo fggvnyre. Eredmnye az f
      nevet, a megoldsok szmt s a futsi idt mp-ben tartalmaz
      fzr.
   megold : string * string -> string
 *)
fun megold (f, m) =
    let val kt = kigyo_be f
	val start = Timer.startCPUTimer()
	val kvk = Kigyo.kigyo kt
	val {usr=tim,...} = Timer.checkCPUTimer start
	val _ = kigyo_ki(m, kt, kvk)
    in
	"Feladvny: " ^ f ^ ", " ^
	Int.toString(length kvk) ^ " megolds, " ^
	"id: " ^ Time.fmt 2 tim ^ " mp\n"
    end

end
